/*
 * Copyright 2010-2017, Tarantool AUTHORS, please see AUTHORS file.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the
 *    following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *
 * This file contains code used to insert the values of host parameters
 * (aka "wildcards") into the SQL text output by sqlite3_trace().
 *
 * The Vdbe parse-tree explainer is also found here.
 */
#include "sqliteInt.h"
#include "vdbeInt.h"

#ifndef SQLITE_OMIT_TRACE

/*
 * zSql is a zero-terminated string of UTF-8 SQL text.  Return the number of
 * bytes in this text up to but excluding the first character in
 * a host parameter.  If the text contains no host parameters, return
 * the total number of bytes in the text.
 */
static int
findNextHostParameter(const char *zSql, int *pnToken)
{
	int tokenType;
	int nTotal = 0;
	int n;
	bool unused;

	*pnToken = 0;
	while (zSql[0]) {
		n = sqlite3GetToken((u8 *) zSql, &tokenType, &unused);
		assert(n > 0 && tokenType != TK_ILLEGAL);
		if (tokenType == TK_VARIABLE) {
			*pnToken = n;
			break;
		}
		nTotal += n;
		zSql += n;
	}
	return nTotal;
}

/*
 * This function returns a pointer to a nul-terminated string in memory
 * obtained from sqlite3DbMalloc(). If sqlite3.nVdbeExec is 1, then the
 * string contains a copy of zRawSql but with host parameters expanded to
 * their current bindings. Or, if sqlite3.nVdbeExec is greater than 1,
 * then the returned string holds a copy of zRawSql with "-- " prepended
 * to each line of text.
 *
 * If the SQLITE_TRACE_SIZE_LIMIT macro is defined to an integer, then
 * then long strings and blobs are truncated to that many bytes.  This
 * can be used to prevent unreasonably large trace strings when dealing
 * with large (multi-megabyte) strings and blobs.
 *
 * The calling function is responsible for making sure the memory returned
 * is eventually freed.
 *
 * ALGORITHM:  Scan the input string looking for host parameters in any of
 * these forms:  ?, ?N, $A, @A, :A.  Take care to avoid text within
 * string literals, quoted identifier names, and comments.  For text forms,
 * the host parameter index is found by scanning the prepared
 * statement for the corresponding OP_Variable opcode.  Once the host
 * parameter index is known, locate the value in p->aVar[].  Then render
 * the value as a literal in place of the host parameter name.
 */
char *
sqlite3VdbeExpandSql(Vdbe * p,	/* The prepared statement being evaluated */
		     const char *zRawSql	/* Raw text of the SQL statement */
    )
{
	sqlite3 *db;		/* The database connection */
	int idx = 0;		/* Index of a host parameter */
	int nextIndex = 1;	/* Index of next ? host parameter */
	int n;			/* Length of a token prefix */
	int nToken;		/* Length of the parameter token */
	int i;			/* Loop counter */
	Mem *pVar;		/* Value of a host parameter */
	StrAccum out;		/* Accumulate the output here */
	char zBase[100];	/* Initial working space */

	db = p->db;
	sqlite3StrAccumInit(&out, 0, zBase, sizeof(zBase),
			    db->aLimit[SQLITE_LIMIT_LENGTH]);
	if (db->nVdbeExec > 1) {
		while (*zRawSql) {
			const char *zStart = zRawSql;
			while (*(zRawSql++) != '\n' && *zRawSql) ;
			sqlite3StrAccumAppend(&out, "-- ", 3);
			assert((zRawSql - zStart) > 0);
			sqlite3StrAccumAppend(&out, zStart,
					      (int)(zRawSql - zStart));
		}
	} else if (p->nVar == 0) {
		sqlite3StrAccumAppend(&out, zRawSql, sqlite3Strlen30(zRawSql));
	} else {
		while (zRawSql[0]) {
			n = findNextHostParameter(zRawSql, &nToken);
			assert(n > 0);
			sqlite3StrAccumAppend(&out, zRawSql, n);
			zRawSql += n;
			assert(zRawSql[0] || nToken == 0);
			if (nToken == 0)
				break;
			if (zRawSql[0] == '?') {
				if (nToken > 1) {
					assert(sqlite3Isdigit(zRawSql[1]));
					sqlite3GetInt32(&zRawSql[1], &idx);
				} else {
					idx = nextIndex;
				}
			} else {
				assert(zRawSql[0] == ':' || zRawSql[0] == '$' ||
				       zRawSql[0] == '@' || zRawSql[0] == '#');
				testcase(zRawSql[0] == ':');
				testcase(zRawSql[0] == '$');
				testcase(zRawSql[0] == '@');
				testcase(zRawSql[0] == '#');
				idx =
				    sqlite3VdbeParameterIndex(p, zRawSql,
							      nToken);
				assert(idx > 0);
			}
			zRawSql += nToken;
			nextIndex = idx + 1;
			assert(idx > 0 && idx <= p->nVar);
			pVar = &p->aVar[idx - 1];
			if (pVar->flags & MEM_Null) {
				sqlite3StrAccumAppend(&out, "NULL", 4);
			} else if (pVar->flags & MEM_Int) {
				sqlite3XPrintf(&out, "%lld", pVar->u.i);
			} else if (pVar->flags & MEM_Real) {
				sqlite3XPrintf(&out, "%!.15g", pVar->u.r);
			} else if (pVar->flags & MEM_Str) {
				int nOut;	/* Number of bytes of the string text to include in output */
				nOut = pVar->n;
#ifdef SQLITE_TRACE_SIZE_LIMIT
				if (nOut > SQLITE_TRACE_SIZE_LIMIT) {
					nOut = SQLITE_TRACE_SIZE_LIMIT;
					while (nOut < pVar->n
					       && (pVar->z[nOut] & 0xc0) ==
					       0x80) {
						nOut++;
					}
				}
#endif
				sqlite3XPrintf(&out, "'%.*q'", nOut, pVar->z);
#ifdef SQLITE_TRACE_SIZE_LIMIT
				if (nOut < pVar->n) {
					sqlite3XPrintf(&out, "/*+%d bytes*/",
						       pVar->n - nOut);
				}
#endif
			} else if (pVar->flags & MEM_Zero) {
				sqlite3XPrintf(&out, "zeroblob(%d)",
					       pVar->u.nZero);
			} else {
				int nOut;	/* Number of bytes of the blob to include in output */
				assert(pVar->flags & MEM_Blob);
				sqlite3StrAccumAppend(&out, "x'", 2);
				nOut = pVar->n;
#ifdef SQLITE_TRACE_SIZE_LIMIT
				if (nOut > SQLITE_TRACE_SIZE_LIMIT)
					nOut = SQLITE_TRACE_SIZE_LIMIT;
#endif
				for (i = 0; i < nOut; i++) {
					sqlite3XPrintf(&out, "%02x",
						       pVar->z[i] & 0xff);
				}
				sqlite3StrAccumAppend(&out, "'", 1);
#ifdef SQLITE_TRACE_SIZE_LIMIT
				if (nOut < pVar->n) {
					sqlite3XPrintf(&out, "/*+%d bytes*/",
						       pVar->n - nOut);
				}
#endif
			}
		}
	}
	if (out.accError)
		sqlite3StrAccumReset(&out);
	return sqlite3StrAccumFinish(&out);
}

#endif				/* #ifndef SQLITE_OMIT_TRACE */
